# AUTOGENERATED! DO NOT EDIT! File to edit: ../02_estimators.ipynb.

# %% auto 0
__all__ = ['CausalForestEstimator', 'OPENBTITEEstimator', 'GPEstimator', 'SKLiftEstimator']

# %% ../02_estimators.ipynb 2
#| output: false
#| output: false
from .base import *
from econml.dml import CausalForestDML
import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression
import econml
import sklift
# import causalml
# import pymc as pm 
# import pymc_bart as pmb

# %% ../02_estimators.ipynb 3
class CausalForestEstimator(BaseITEEstimator):
    def fit(self, **kwargs):
        if self.model is None:
            self.model = CausalForestDML()
        self.model.fit(Y=kwargs["y_training"],
                       T=kwargs["t_training"],
                       X=kwargs["X_training"])

    def predict(self, **kwargs):
        if 'return_counterfactuals' in kwargs:
            raise ValueError("Causal Forest does not support counterfactual predictions out of the box")
        preds = self.model.effect_inference(kwargs["X"])
        if "return_mean" in kwargs:
            out = preds.pred
        else:
            out = (preds.pred, preds.var)
        return out

# %% ../02_estimators.ipynb 4
class OPENBTITEEstimator(BaseITEEstimator):
    """Modified ITE estimator for OPENBT

    The predictions are transposed so the uncertainty sampler can calculate uncertianty easily"""
    def predict(self, **kwargs):
        X = kwargs["X"]
        if self.ps_model is not None:
            ps_scores = self.ps_model.predict_proba(X)
            X = np.hstack((X, ps_scores[:,1].reshape((-1, 1))))
        X0 = np.concatenate((X,
                             np.zeros(X.shape[0]).reshape((-1,1))),axis=1)
        X1 = np.concatenate((X,
                             np.ones(X.shape[0]).reshape((-1,1))),axis=1)
        preds0 = self.model.predict(X0)
        preds1 = self.model.predict(X1)
        if "return_mean" in kwargs:
            if kwargs["return_mean"]:
                out = preds1["mmean"] - preds0["mmean"]
        else:
            out = preds1["mdraws"].T - preds0["mdraws"].T
        if "return_per_cf" in kwargs:
            if kwargs["return_per_cf"]:
                return {"pred1": preds1["mdraws"].T , "pred0":preds0["mdraws"].T}
        return out

# %% ../02_estimators.ipynb 6
class GPEstimator(BaseITEEstimator):
    def predict(self, **kwargs):
        if 'return_mean' in kwargs:
            pred0 = self.model.predict(kwargs["X"])
            pred1 = self.m1.predict(kwargs["X"])
            ite = pred1 - pred0
        else:
            draws0 = self.model.sample_y(kwargs["X"], n_samples=100)
            draws1 = self.m1.sample_y(kwargs["X"], n_samples=100)
            ite = draws1 - draws0
        return ite

# %% ../02_estimators.ipynb 8
class SKLiftEstimator(BaseITEEstimator):
    def __init__(self, sk_model):
        self.model = sk_model
    
    def fit(self, **kwargs):
        self.model.fit(X=kwargs["X_training"],
                       y=kwargs["y_training"],
                       treatment=kwargs["t_training"])
    
    def predict(self, **kwargs):
        if 'return_counterfactuals' in kwargs:
            raise ValueError("Causal Forest does not support counterfactual predictions out of the box")
        out = self.model.predict(kwargs["X_test"])
        return out



# AUTOGENERATED! DO NOT EDIT! File to edit: ../99_misc.ipynb.

# %% auto 0
__all__ = ['get_ihdp_dict']

# %% ../99_misc.ipynb 2
from .base import *
from .models import *
from .estimators import *
import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from copy import deepcopy

# %% ../99_misc.ipynb 3
def get_ihdp_dict(i = 1, test_size = 0.9, seperate_pool_test = False, pool_size=0.8, seed = None):
    """
    Method to query the IHDP data from the AMlab github repo

    Arguments
    ---------
    i : int = 1
        Which IHDP dataset to get (between 1 and 10)
    test_size : 0.9
        Fraction of the data used for test set
    seperate_pool_test:
        Whether the pool and the test set are different
    pool_size :
        Fraction of test data when pool vs test split is done
    seed :
        Seed when splitting data
    """
    seed = 1005 if seed is None else seed
    df = pd.read_csv(
    f"https://raw.githubusercontent.com/AMLab-Amsterdam/CEVAE/master/datasets/IHDP/csv/ihdp_npci_{i}.csv",
    names = ["treatment", "y_factual", "y_cfactual", "mu0", "mu1"] + [f'x{x}' for x in range(25)])

    X = df.loc[:,"x0":].to_numpy()
    t = df["treatment"].to_numpy()
    #t = np.zeros_like(t)
    y = df["y_factual"].to_numpy()
    y1 = np.where(df["treatment"] == 1,
                   df['y_factual'],
                   df['y_cfactual'])
    y0 = np.where(df["treatment"] == 0,
                   df['y_factual'],
                   df['y_cfactual'])
    ite = np.where(df["treatment"] == 1,
                   df['y_factual'] - df["y_cfactual"],
                   df['y_cfactual'] - df["y_factual"])
    if seperate_pool_test:
        X_train, X_pool, t_train, t_pool, y_train, y_pool, ite_train, ite_pool, y1_train, y1_pool, y0_train, y0_pool = train_test_split(
    X, t, y, ite, y1, y0,  test_size=test_size, random_state=seed)
        X_pool, X_test, t_pool, t_test, y_pool,y_test, ite_pool, ite_test, y1_pool,y1_test, y0_pool, y0_test = train_test_split(
    X_pool, t_pool, y_pool, ite_pool, y1_pool, y0_pool, test_size=pool_size, random_state=seed)
        ds = {"X_training": X_train,
             "y_training": y_train,
             "t_training": t_train,
             "X_pool": X_pool,
             "y_pool": y_pool,
             "t_pool": t_pool,
             "y1_pool": y1_pool,
             "y0_pool":y0_pool,
             "X_test": X_test,
             "y_test": y_test,
              "t_test": t_test,
              "ite_training" : ite_train,
              "ite_pool" : ite_pool,
              "ite_test": ite_test
             }
    else:
        X_train, X_test, t_train, t_test, y_train, y_test, ite_train, ite_test, y1_train, y1_test, y0_train, y0_test = train_test_split(
    X, t, y, ite, y1, y0,  test_size=test_size, random_state=1005)
        ds = {"X_training": X_train,
     "y_training": y_train,
     "t_training": t_train,
     "X_pool": deepcopy(X_test),
     "y_pool": deepcopy(y_test),
     "t_pool": deepcopy(t_test),
     "y1_pool": y1_test,
     "y0_pool":y0_test,
     "X_test": X_test,
     "y_test": y_test,
      "t_test": t_test,
      "ite_test": ite_test
     }
    return ds



# AUTOGENERATED! DO NOT EDIT! File to edit: ../00_base.ipynb.

# %% auto 0
__all__ = ['FitTask', 'BaseITEEstimator', 'BaseActiveLearner', 'BaseAcquisitionFunction', 'BaseAssignmentFunction',
           'BaseStoppingRule', 'BaseDataGenerator']

# %% ../00_base.ipynb 2
#| output: false
from sklearn.base import BaseEstimator
import numpy as np
from typing import Union, Callable, Optional, Tuple, List, Iterator, Any
from copy import deepcopy
from dataclasses import dataclass, field
#from sklift.metrics import qini_auc_score, qini_curve
#from pylift.eval import UpliftEval
from pylift.eval import get_scores
from fastcore.test import *
import asbe

# %% ../00_base.ipynb 4
class FitTask(type):
    """Meta class to make preprocessing of data before fitting different models

    This meta class is needed so new models can be incorporated easily
    by only having a .fit() method. It will call a prepare_data function, which
    can also be modified .
    """
    # https://stackoverflow.com/a/18859019
    def __init__(cls, name, bases, clsdict):
        if 'fit' in clsdict:
            def new_fit(self, **kwargs):
                try:
                    X_training = kwargs["X_training"] if "X_training" in kwargs else self.dataset["X_training"]
                    y_training = kwargs["y_training"] if "y_training" in kwargs else self.dataset["y_training"]
                    t_training = kwargs["t_training"] if "t_training" in kwargs else self.dataset["t_training"]
                except:
                    raise ValueError("Can't find (X,t,y) data to fit the model on")
                try:
                    ps_scores  = kwargs["ps_scores"] if "ps_scores" in kwargs else self.dataset["ps_scores"]
                except:
                    ps_scores = None
                try:
                    fun_to_prepare = bases['prepare_data'] if ["prepare_data"] in dir(
                        bases) else self.prepare_data
                    X_training, t_training, y_training, ps_scores = fun_to_prepare(X_training, t_training, y_training, ps_scores)
                except:
                    raise ValueError("Can't prepare data. Use either the default or implement new 'prepare_data' method")
                clsdict['fit'](self, X_training = X_training,
                               t_training = t_training,
                               y_training = y_training,
                               ps_scores = ps_scores)
            setattr(cls, 'fit', new_fit)

# %% ../00_base.ipynb 5
class BaseITEEstimator(BaseEstimator, metaclass=FitTask):
    """Base class for estimating treatment effects

    This is a base class, that is also usable on its own to estimate treatment effects.
    It works with most models that have a .fit() method, but subclassing is also made
    straightforward throught a metaclass.

    Attributes
    ----------
    model : str, Callable
        treatment effect estimator to be used
    dataset: dict, BaseDataGenerator
        dataset to be used by the model, either offline or online
    two_model: bool
        Switches between S or T learner approach, when appropriate
    ps_model: Callable, None
        Propensity score model, if used, must be classifier

    Methods
    -------
    prepare_data(X_training, t_training, y_training, ps_scores):
        Puts together data for training before .fit() is called

    fit(X_training, t_training, y_training, ps_scores):
        Fits the treatment effect estimator to the training data

    predict(X, ps_scores):
        Predicts the treatment effects based on the supplied X
    """

    def __init__(self,
                model: Union[str, Callable] = None,
                dataset : Union[dict,None] = None,
                two_model: Union[bool,None] = None,
                ps_model : Union[Callable, None] = None,
                **kwargs)-> None:
        self.model = model
        self.dataset = dataset
        self.two_model = two_model
        self.ps_model = ps_model
        self.proba = False
        """
        Makes the estimator ready for its task

        Parameters
        ----------
        model : str, Callable
            treatment effect estimator to be used
        dataset: dict, BaseDataGenerator
            dataset to be used by the model, either offline or online
        two_model: bool
            Switches between S or T learner approach, when appropriate
        ps_model: Callable, None
            Propensity score model, if used, must be classifier

        Returns
        -------
        None
        """

    def _fit_ps_model(self, X_training, t_training):
        self.ps_model.fit(X_training, t_training)

    def prepare_data(self, X_training=None, t_training=None, y_training=None, ps_scores=None):
        """
        Prepares data before fitting the model

        Parameters
        ----------
        X_training : np.ndarray
            Training features (size n x d)
        t_training : np.ndarray
            Training treatments (size n x 1)
        y_training : np.ndarray
            Training labels (size n x 1)
        ps_scores : np.ndarray, None
            Optinal propensity scores if they are coming from outside model

        Returns
        -------
        X_training, t_training, y_training, ps_scores
            Updated training data
        """
        if X_training is None:
            try:
                X_training = deepcopy(self.dataset["X_training"])
                y_training = deepcopy(self.dataset["y_training"])
                t_training = deepcopy(self.dataset["t_training"])
            except:
                raise ValueError("No data to fit the model on.")
        if np.unique(y_training).shape[0] == 2:
            self.proba = True
        if self.ps_model is not None or ps_scores is not None:
            if self.ps_model is not None:
                self._fit_ps_model(X_training, t_training)
                ps_scores = self.ps_model.predict_proba(X_training)
            try:
                X_training = np.hstack((X_training, ps_scores[:,1].reshape((-1, 1))))
            except:
                raise ValueError(f"Shape of propensity scores is {ps_scores.shape},instead of (-1,1)")
        X_t_training = np.hstack((X_training, t_training.reshape(-1,1)))
        if self.two_model is False:
            X_training = deepcopy(X_t_training)

        return X_training, t_training, y_training, ps_scores

    def fit(self,
            X_training,
            t_training,
            y_training,
            ps_scores,
            **kwargs):
        """
        Fits the model using .fit() function of other the given ITE estimators

        Parameters
        ----------
        X_training : np.ndarray
            Training features (size n x d)
        t_training : np.ndarray
            Training treatments (size n x 1)
        y_training : np.ndarray
            Training labels (size n x 1)
        ps_scores : np.ndarray, None
            Optinal propensity scores if they are coming from outside model

        Returns
        -------
        None
        """
        if self.two_model is False or self.two_model is None:
            self.model.fit(X_training, y_training)
        else:
            control_ix = np.where(t_training == 0)[0]
            self.m1 = deepcopy(self.model)
            self.model.fit(X_training[control_ix,:], y_training[control_ix])
            self.m1.fit(X_training[-control_ix,:],   y_training[-control_ix])

    def _predict_bin_or_con(self, model, X):
        if self.proba:
            try:
                out = model.predict_proba(X)[:, 1]
            except:
                out = model.predict(X)
        else:
            out = model.predict(X)
        return out

    def predict(self,
                    X: np.ndarray,
                    ps_scores = None,
                    **kwargs):
        """
        Predicts the treatment effect using the fitted model on the given features

        Parameters
        ----------
        X : np.ndarray
            Features (size n x d)
        ps_scores : np.ndarray, None
            Optinal propensity scores if they are coming from outside model

        Returns
        -------
        Predicted treatment effect scores (size n x 1 / n x B)
        """
        if self.ps_model is not None or ps_scores is not None:
            if self.ps_model is not None:
                ps_scores = self.ps_model.predict_proba(X)
            X = np.hstack((X, ps_scores[:,1].reshape((-1, 1))))
        if self.two_model:
            m0_preds = self._predict_bin_or_con(self.model, X)
            m1_preds =  self._predict_bin_or_con(self.m1, X)
            ite = m1_preds - m0_preds
        else:
            try:
                X0 = np.hstack((X, np.zeros((X.shape[0], 1))))
                X1 = np.hstack((X, np.ones((X.shape[0], 1))))
                m1_preds = self._predict_bin_or_con(self.model, X1)
                m0_preds = self._predict_bin_or_con(self.model, X0)
                ite = m1_preds - m0_preds
            except:
                ite = self._predict_bin_or_con(self.model, X)
        return np.asarray(ite)

# %% ../00_base.ipynb 7
class BaseActiveLearner(BaseEstimator):
    """Basic of Active Learners later used, with the capability for treatment effects

    Inspired by modAL's BaseLearner, however modified for ITE estimation. Dataset is provided by
    a dictionary for easier usage and less clutter, but alternatively a BaseDataGenerator can
    also be supplied.

    Attributes
    ----------
    estimator: BaseEstimator
        Estimator for treatment effects, which expects a BaseEstimator or a subclass of it
    acquisition_function:
        Method that selects units to label
    assignment_function:
        Method that selects control or treatment for selected units
    stopping_function:
        Method that determines if the active learning process should stop after the current step
    dataset : dict, BaseDataGenerator
        Dataset that is used for the active learning process
    al_steps : int = 1
        number of active learnign steps to take. Can be controlled from the outside by self.current_step
    offline : bool
        Determines whether data is supplied beforehand or created on the fly

    Methods
    -------
    fit():
        Fits the supplied estimator with data from dataset
    predict(X):
        Predicts X with the estimator
    query(no_query = None, acquisition_function = None):
        Calls the acquisition function to query datapoints to label.
        Number of queries and AF can be supplied, which overwrited the active learner's
        at the current step
    teach(query_idx=None, assignment_function = None, **kwargs):
        Teaches the supplied indices, by moving them to the training set and
        selects the counterfactuals used
    score(metric="PEHE"):
        Uses a test set to calculate a metric from the estimator
    simulate(no_query: int = None, metric: str = "Qini")
        Simulates the active learning process by going through al_steps.
        The metric for the simulation to be saved can be set
    plot():
        Plots the typical active learning style line plot, after simulation
        has finished.
    """
    def __init__(self,
                 estimator: BaseEstimator,
                 acquisition_function: Callable,
                 assignment_function: Union[Callable, list, None],
                 stopping_function: Union[Callable, list, None],
                 dataset: dict,
                 al_steps: int = 1,
                 offline = True,
                 **kwargs
                ) -> None:
        """Initiates the active learning sequence

        If the learning happens offline, sets up calls to the data generating processes

        Parameters
        ----------
        estimator: BaseEstimator
            Estimator for treatment effects, which expects a BaseEstimator or a subclass of it
        acquisition_function:
            Method that selects units to label
        assignment_function:
            Method that selects control or treatment for selected units
        stopping_function:
            Method that determines if the active learning process should stop after the current step
        dataset : dict, BaseDataGenerator
            Dataset that is used for the active learning process
        al_steps : int = 1
            number of active learnign steps to take. Can be controlled from the outside by self.current_step
        offline : bool
            Determines whether data is supplied beforehand or created on the fly

        Returns
        -------
        None
        """
        self.estimator = estimator
        if type(acquisition_function) is list:
            self.acquisition_function_list = acquisition_function
            self.acquisition_function = None
        else:
            self.acquisition_function = acquisition_function
            self.acquisition_function_list = []
        if type(assignment_function) is list:
            self.assignment_function = None
            self.assignment_function_list = assignment_function
        else:
            self.assignment_function = assignment_function
            self.assignment_function_list = []
        self.stopping_function = stopping_function
        self.dataset = deepcopy(dataset)
        self.al_steps = al_steps
        self.current_step = 0
        self.next_query = True
        self.offline = offline
        if self.offline is False:
            self.dataset.__dict__["X_training"],\
            self.dataset.__dict__["t_training"],\
            self.dataset.__dict__["y_training"] = self.dataset.get_data(
                self.dataset.no_training)


    def _update_dataset(self, query_idx, **kwargs):
        """Moves data with query_idx indices from pool to training"""
        remove_data = kwargs["remove_data"] if "remove_data" in kwargs else True
        if type(self.dataset) is not dict:
            self.dataset = self.dataset.__dict__
        if "outside_data" in kwargs:
            for data in ["X", "t", "y"]:
                self.dataset[f"{data}_training"] = np.concatenate((
                            self.dataset[f"{data}_training"],
                            kwargs["outside_data"][f"{data}"]))
            return None
        data_available = list(set([x.split("_")[0] for x in self.dataset.keys()]))
        for data in ["X", "t", "y", "ite"]:
            if data in data_available and f"{data}_training" in self.dataset.keys():
                if self.dataset[f"{data}_training"].shape[0] > 0 :
                    if data in ["X"]:
                        self.dataset[f"{data}_training"] = np.concatenate((
                            self.dataset[f"{data}_training"],
                            self.dataset[f"{data}_pool"][query_idx,:]))
                    else:
                        self.dataset[f"{data}_training"] = np.concatenate((
                            self.dataset[f"{data}_training"],
                            self.dataset[f"{data}_pool"][query_idx]))
                else:
                    self.dataset[f"{data}_training"] = self.dataset[f"{data}_pool"][query_idx,:]
                if remove_data:
                    self.dataset[f"{data}_pool"] = np.delete(self.dataset[f"{data}_pool"],
                                                                 query_idx, 0)
        try:
            for cfs in ["y0","y1"]:
                self.dataset[f"{cfs}_training"] = np.concatenate((
                                self.dataset[f"{cfs}_training"],
                                self.dataset[f"{cfs}_pool"][query_idx]))
                self.dataset[f"{cfs}_pool"] = np.delete(self.dataset[f"{cfs}_pool"],
                                                                 query_idx, 0)
        except KeyError:
            pass
        return None


    def _select_counterfactuals(self, query_idx, treatment):
        "Rewrites observed outcomes in pool for counterfactuals"
        self.dataset["y_pool"][query_idx] = np.where(treatment == 1,
                                                     self.dataset["y1_pool"][query_idx],
                                                     self.dataset["y0_pool"][query_idx])

        self.dataset["t_pool"][query_idx] = treatment

    def fit(self) -> None:
        """
        Fits the model using the given estimator
        """
        self.estimator.fit(X_training = self.dataset["X_training"],
                           t_training = self.dataset["t_training"],
                           y_training = self.dataset["y_training"])
        return None

    def predict(self, X):
        """
        Predicts using the the given estimator

        Attributes
        ----------
        X : np.ndarray
            The features to predict on

        Returns
        -------
        Predicted treatment effects
        """
        return self.estimator.predict(X=X)

    def query(self, no_query = None,
              acquisition_function = None,
              return_all=False,
              **kwargs):
        """Main function to select datapoints for labeling

        Calls the acquisition function to determine which units to label.
        The acquisition function can be overwritten for a given query and the number
        of queries as well. Queries happen until the budget is exhauested, which is
        controlled by the stopping function.

        Parameters
        ----------
        no_query : int = None
            Number of queries
        acquisition_function
            Optional acquisition function
        return_all
            Whether to return a dict of X,t,y or only indices + X

        Returns
        -------
        X_new, query_idx :
            Selected features, selected indices
        """
        if no_query is None:
            no_query = self.no_query
        if self.stopping_function is not None:
            self.next_query = self.stopping_function.check_rule(
                self.estimator, self.dataset, self.step)
        if self.next_query:
            if len(self.acquisition_function_list) == 0:
                if acquisition_function is None:
                    acquisition_function = self.acquisition_function
            else:
                if acquisition_function is None:
                    acquisition_function = self.acquisition_function
            if self.offline:
                X_new, query_idx = acquisition_function.select_data(self.estimator,
                                                              self.dataset,
                                                              no_query,
                                                              **kwargs)
            else:
                X_get = self.dataset.get_X(no_query = no_query)
                data_to_estimate = {"X_training":self.dataset["X_training"],
                                   "X_pool": X_get}
                decision_to_query = acquisition_function.select_data(self.estimator,
                                                              data_to_estimate,
                                                              no_query,
                                                              offline = False,
                                                              **kwargs)
                if decision_to_query:
                    self.X_to_add = X_get
                    X_new = X_get
                else:
                    X_new = None
                try:
                    query_idx = self.dataset.selected_ix
                except:
                    query_idx = None
        else:
            X_new, query_idx = None, None
        if return_all:
            return {"X": X_new,
                    "t": self.dataset["t_pool"][query_idx],
                    "y": self.dataset["y_pool"][query_idx]}
        return X_new, query_idx

    def teach(self, query_idx=None, assignment_function = None, select_again = False, **kwargs):
        """Function to assign the selected labels to the training set

        Selects counterfactuals (if possible), through the assignmnet function
        and moves data from the pool to the training set.

        Parameters
        ----------
        query_idx : np.array = None
            the indices of queryed samples
        assignment_function : None
            Optional assignment function to be used to select counterfactuals
        select_again = False
            If there are no matching, select units until all are matching
        kwargs["outside_data"] = None
            If outside data is given in a dictionary, the assignment function won't do anything
            it will assign the data given in this dict

        Returns
        -------
        None
        """
        if "outside_data" in kwargs:
            self._update_dataset(query_idx = None,
                                 outside_data = kwargs["outside_data"])
            return None # exit update immediately
        
        if len(self.assignment_function_list) == 0:
            if assignment_function is None:
                if self.assignment_function is None:
                    self.assignment_function = asbe.models.RandomAssignmentFunction()
                assignment_function = self.assignment_function
            else:
                raise ValueError("No assignment function provided")
        else:
            if assignment_function is None:
                assignment_function = self.assignment_function
        if self.offline:
            treatment = assignment_function.select_treatment(self.estimator,
                                                              self.dataset,
                                                              query_idx)
            if "y0_pool" in self.dataset:
                self._select_counterfactuals(query_idx, treatment)
            matching = treatment == self.dataset["t_pool"][query_idx]
            if matching.all():
                self._update_dataset(query_idx, **kwargs)
            else:
                self._update_dataset(query_idx[matching], **kwargs)
                if select_again:
                    while matching.sum() < treatment.shape[0]:
                        X_reselected, ix_reselected = self.query(
                            treatment.shape[0] - matching.sum())
                        treatment = assignment_function.select_treatment(self.estimator,
                                                              self.dataset,
                                                              ix_reselected)
                        matching = treatment == self.dataset["t_pool"][ix_reselected]
                        self._update_dataset(ix_reselected[matching], **kwargs)
        else:
            treatment_to_add = self.dataset.get_t(X_new = self.X_to_add)
            y = self.dataset.get_y(X_new=self.X_to_add, t_new = treatment_to_add)
            self.dataset.__dict__["X_pool"] = self.X_to_add
            self.dataset.__dict__["t_pool"] = treatment_to_add
            self.dataset.__dict__["y_pool"] = y
            self._update_dataset(np.arange(y.shape[0]), remove_data=True)

        return None

    def score(self, metric="PEHE"):
        """
        Calculates the metric given on the test set.

        Parameters
        ----------
        metric : str = "PEHE"
            Metric to be used when calculating the performance

        Returns
        -------
        The score of the model
        """
        metrics = ["Qini", "PEHE", "Cgains", "decision", "Qini_curve"]
        if not callable(metric):
            if metric not in metrics:
                raise ValueError(f"Please use a valid error ({metrics}), {metric} is not valid")
        try:
            preds = self.estimator.predict(X=self.dataset["X_test"], return_mean=True)
        except:
            preds = self.estimator.predict(X=self.dataset["X_test"])
        if metric == "PEHE":
            try:
                sc = np.sqrt(np.mean(np.square(preds - self.dataset["ite_test"])))
            except KeyError:
                raise Error("Check if dataset contains true ITE values")
        elif metric == "decision":
            dec = np.where((preds >= 0) &( self.dataset["ite_test"] >= 0), 1, 0)
            sc = np.sum(dec)/self.dataset["ite_test"].shape[0]
        elif metric == "Qini":
            sc = get_scores(self.dataset["t_test"],
                            self.dataset["y_test"],
                            preds,
                            np.zeros_like(self.dataset["t_test"]) + self.dataset["t_test"].mean())["q1_qini"]
#         elif metric == "Qini_curve":
#             sc = qini_curve(y_true=self.dataset["y_test"],
#                                 uplift=preds,
#                                 treatment=self.dataset["t_test"])
        elif callable(metric):
            try:
                sc = metric(preds, self.dataset)
            except:
                raise ValueError("Metric can't be used with current data")
        return sc

    def simulate(self, no_query: int = None, metric: str = "Qini") -> dict:
        """
        Simulates the active learning process based on the attributes to the class

        Parameters
        ----------
        no_query:
            Number of units to be queried
        Metric:
            Metric to be calculated at the end of each active learning step

        Returns
        -------
        dict
            Dictionary with keys of assignment functions and values of steps and scores
        """
        ds = deepcopy(self.dataset)
        est = deepcopy(self.estimator)
        res = {}
        if len(self.acquisition_function_list) == 0:
            laf = [self.acquisition_function]
        else:
            laf = self.acquisition_function_list
        for af in laf:
            if no_query is None:
                no_query = af.no_query
            self.dataset = deepcopy(ds)
            self.estimator = deepcopy(est)
            self.current_step = 0
            res[af.name] = {}
            if type(metric) is list:
                for m in metric:
                    res[af.name][m]={}
            for i in range(1, self.al_steps+1):
                self.fit()
                X_new, query_idx = self.query(no_query=no_query, acquisition_function = af)
                if self.next_query:
                    if self.offline:
                        self.teach(query_idx)
                    else:
                        self.teach()
                    if type(metric) is list:
                        for m in metric:
                            res[af.name][m][i] = self.score(metric=m)
                    else:
                        res[af.name][i] = self.score(metric=metric)
                    self.current_step += 1
        self.simulation_results = res
        return res

    def plot(self):
        """Plots the results of the AL simulation"""
        pd.DataFrame(self.simulation_results).plot()

# %% ../00_base.ipynb 8
class BaseAcquisitionFunction():
    """Base class for acquisition functions

    Based on Eq. 3 in paper:
        - calculate_metrics is used to get infromativeness/representativeness
        - calculate_metrics returns a weighted average between the two
        - select_data does the normalization and selecting top m/Bernoulli draws

    Attributes
    ----------
    no_query : int = 1
        Number of queries
    method : str = "top"
        Way to select the data after metrics are calculated
    name : str
        Name of the AF, used in simulations

    Methods
    -------
    calculate_metrics(model, dataset):
        User defined method to get different metrics about the pool set
    select_data(model, dataset, no_query):
        Based on the calculated metrics, selects the data to be added to the
        training set.
    """
    def __init__(self,
                no_query: int = 1,
                method: str = "top",
                name: str = "base") -> None:
        self.no_query = no_query
        self.method = method
        self.name = name + "_" + str(no_query)

    def calculate_metrics(self, model, dataset) -> np.array:
        """
        Method to calculate base metrics for assignment function

        Parameters
        ----------
        model:
            The ITE estimator
        dataset:
            The dataset dictionary/class that holds the training/pool set.

        Returns
        -------
        weighted average of individual metrics
        """
        return np.arange(dataset["X_pool"].shape[0]) + 1

    def select_data(self, model, dataset, no_query, offline=True, **kwargs):
        """
        Based on the calculated metrics, select data and return X and indexes

        Parameters
        ----------
        model:
            The ITE estimator
        dataset:
            The dataset dictionary/class that holds the training/pool set.
        no_query:
            Number of queries
        offline : bool
            Indicator whether the process is online/offline

        Returns
        -------
        Tuple of new features and indices
        """
        if no_query is None:
            no_query = self.no_query
        if dataset["X_training"].shape[0] == 0:
            try:
                metrics = self.calculate_metrics(model, dataset, **kwargs)
            except:
                metrics = np.random.shuffle(np.arange(dataset["X_pool"].shape[0]))
        else:
            metrics = self.calculate_metrics(model, dataset, **kwargs)

        if offline:
            if self.method in ["top", "normalized", "only_ones"]:
                if self.method == "top":
                    query_idx = np.argsort(np.asarray(metrics))[-no_query:][::-1]
                elif self.method == "normalized":
                    p_j = (metrics - np.min(metrics))/(np.max(metrics) - np.min(metrics))
                    query_idx = []
                    for pix in p_j.argsort()[::-1]:
                        if np.random.binomial(1, p = p_j[pix]) == 1:
                            query_idx.append(pix)
                        if len(query_idx) == no_query:
                            break
                    query_idx = np.asarray(query_idx)
                elif self.method == "only_ones":
                    query_idx = np.where(np.asarray(metrics) == 1)[0]
                    query_idx = query_idx.tolist()
            else:
                raise NotImplementedError("Please use a method that is implemented")
            X_new = dataset["X_pool"][query_idx,:]
            out = (X_new, query_idx)
        else:
            out = metrics.sum() >= metrics.shape[0]/2
        return out

# %% ../00_base.ipynb 9
class BaseAssignmentFunction():
    """Base class for assignment functions

    Arguments
    ---------
    base_selection : int = 0
        Which treatment to select automatically

    Methods
    -------
    select_treatment(model, dataset, query_idx)
        Selects the treaatment based on the model and the dataset.
        Predictions can only be done for the selected indices to save time.
    """

    def __init__(self, base_selection = 0):
        self.base_selection = base_selection

    def select_treatment(self, model, dataset, query_idx):
        return dataset["t_pool"][query_idx]

# %% ../00_base.ipynb 10
class BaseStoppingRule():
    """Base class for providing a stopping rule for the active learner

    Arguments
    ---------
    budget : int = None
        The budget that the active learner can use to run simulations

    Methods
    -------
    check_rules(model, dataset, step)
        The method to check if the condition is reached to stop the active
        learning loop
    """

    def __init__(self, budget=None):
        self.budget = budget

    def check_rule(self, model, dataset, step):
        if self.budget - step >= 0:
            out = False
        else:
            out = True
        return out

# %% ../00_base.ipynb 11
@dataclass
class BaseDataGenerator():
    """Class to generate online data for active learing purposes"""
    ds: Union[dict, Callable]
    dgp_x : Union[Callable, None] = None
    dgp_t : Union[Callable, None] = None
    dgp_y : Union[Callable, None] = None
    no_training : int = 10

    def __getitem__(self, key):
        return super().__getattribute__(key)

    def get_X(self, method : str = "random", no_query : int = 1):
        if self.ds is not None:
            ix = np.random.randint(low = 0, high = self.ds["X_pool"].shape[0], size=(no_query))
            X_new = self.ds["X_pool"][ix,:]
            self.selected_ix = ix
        else:
            X_new = self.dgp_x(no_query)
        self.X_new = X_new
        return X_new

    def get_t(self, X_new=None):
        if X_new is None:
            X_new = self.X_new
        if self.ds is not None:
            t_new = self.ds["t_pool"][self.selected_ix]
        else:
            t_new = self.dgp_t(X_new)
        self.t_new = t_new
        return t_new

    def get_y(self, X_new=None, t_new=None):
        if X_new is None:
            X_new = self.X_new
        if t_new is None:
            t_new = self.t_new
        if self.ds is not None:
            y_new = np.where(t_new == 1,
                             self.ds["y1_pool"][self.selected_ix],
                             self.ds["y0_pool"][self.selected_ix])
        else:
            y_new = self.dgp_y(X_new, t_new)
        return y_new

    def get_data(self, no_query=1, as_test=False):
        X_new = self.get_X(no_query=no_query)
        t_new = self.get_t()
        y_new = self.get_y()
        if as_test:
            self.X_test = X_new
            self.t_test = t_new
            self.y_test = y_new
        return (X_new, t_new, y_new)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../01_models.ipynb.

# %% auto 0
__all__ = ['RandomAcquisitionFunction', 'UncertaintyAcquisitionFunction', 'TypeSAcquistionFunction', 'EMCMAcquisitionFunction',
           'RandomAssignmentFunction', 'UncertaintyAssignmentFunction', 'MajorityAssignmentFunction',
           'PMajorityAssignmentFunction']

# %% ../01_models.ipynb 3
#| output: false
#| output: false
import numpy as np
from .base import *
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.linear_model import SGDRegressor
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import make_pipeline
from typing import Union, Optional, Callable
from copy import deepcopy

# %% ../01_models.ipynb 5
class RandomAcquisitionFunction(BaseAcquisitionFunction):
    def calculate_metrics(self, model, dataset):
        ixs = np.arange(dataset["X_pool"].shape[0])
        np.random.shuffle(ixs)
        return ixs

class UncertaintyAcquisitionFunction(BaseAcquisitionFunction):
    def calculate_metrics(self, model, dataset):
        preds = model.predict(X = dataset["X_pool"])
        if type(preds) is tuple:
            p, pred_var = preds[0], preds[1]
        try:
            if preds.shape[1] <= 1:
                raise Exception("Not possible to calculate uncertainty when dimensions <=1")
            pred_var = np.var(preds, axis = 1)
        except IndexError:
            raise Exception("Not possible to calculate uncertainty when dimensions <=1")
        return pred_var

class TypeSAcquistionFunction(BaseAcquisitionFunction):
    def calculate_metrics(self, model, dataset):
        preds = model.predict(X=dataset["X_pool"])
        if preds.shape[0] <= 1:
            raise Exception("Type S error needs multiple values per prediction")
        prob_s = np.sum(preds > 0, axis=1)/preds.shape[1]
        prob_s_sel = np.where(prob_s > 0.5, 1-prob_s, prob_s) + .0001
        return prob_s_sel

class EMCMAcquisitionFunction(BaseAcquisitionFunction):
    def __init__(self, no_query = 1,
                 method = "top",
                 name = "emcm",
                 approx_model = SGDRegressor(),
                 B = 100,
                 K = 5,
                 threshold = 0,
                 k_means=False,
                 k_mean_clusters=1):
        super().__init__(no_query, method, name)
        self.approx_model = approx_model
        self.B = B
        self.K = K
        self.threshold = threshold
        self.model_change = []
        self.k_means = k_means
        self.k_means_clusters = k_mean_clusters

    def calculate_metrics(self, model, dataset, **kwargs):
        ite_train_preds = model.predict(X = dataset["X_training"], training = True)
        ite_pool_preds = model.predict(X = dataset["X_pool"])
        if ite_train_preds.shape[1] < 1:
            raise ValueError("The treatment effect does not have uncertainty around it - \
                         consider using a different estimator")
        sc = StandardScaler()
        X_scaled = sc.fit_transform(dataset["X_training"])
        # Fit approx model
        # calc type-s error
        train_type_s_prob_1 = np.sum(ite_train_preds > 0, axis=1)/ite_train_preds.shape[1]
        train_type_s = np.where(train_type_s_prob_1 > 0.5, 1-train_type_s_prob_1, train_type_s_prob_1) + .0001
        pool_type_s_prob_1 = np.sum(ite_pool_preds > 0, axis=1)/ite_pool_preds.shape[1]
        pool_type_s = np.where(pool_type_s_prob_1 > 0.5, 1-pool_type_s_prob_1, pool_type_s_prob_1) + .0001
        self.approx_model.fit(
            X = X_scaled,
            y = np.mean(ite_train_preds, axis=1),
            sample_weight = self.K*train_type_s)
        # Using list as it is faster than appending to np array
        query_idx = []
        # Using a loop for the combinatorial opt. part
        for ix in range(self.no_query):
            if self.no_query > (dataset["X_pool"].shape[0]):
                raise IndexError("Too many samples are queried from the pool ($n_2 > ||X_pool||$)")
            # Select randomly from X_pool
            prob_sampling = np.ones((dataset["X_pool"].shape[0]))/(
                dataset["X_pool"].shape[0]-len(query_idx))
            # Set the probability of already selected samples to 0
            if ix > 0:
                prob_sampling[query_idx] = 0
            # B = 100 by default, can be modified by kwargs
            considered_ixes = np.random.choice(dataset["X_pool"].shape[0],
                                             size = self.B,
                                             replace=False,
                                             p=prob_sampling)
            # Calculate the grads for all
            grads = np.array([])
            for considered_ix in considered_ixes:
                new_X = sc.transform(dataset["X_pool"][considered_ix].reshape(1, -1))
                app_predicted_ite = self.approx_model.predict(new_X)
                # bootstrapping accroding to eq. 11 of Cai and Zhang
                true_ite = np.random.choice(ite_pool_preds[considered_ix],
                                            size=self.B)
                grad = np.sum(np.abs(np.kron((true_ite - app_predicted_ite),new_X)))
                grads = np.append(grads, grad)
            # Setting up stopping rule after 10% of data has been seen
            if self.threshold != 0:
                print(ix)
                if ix > .1*no_query:
                    print(np.max(grads))
                    print(self.threshold * np.mean(self.model_change))
                    if np.max(grads) < self.threshold * np.mean(self.model_change):
                        break
            self.model_change = np.append(self.model_change,np.max(grads))
            query_idx.append(int(considered_ixes[np.argmax(grads)]))
            self.approx_model.partial_fit(
                sc.transform(dataset["X_pool"][int(query_idx[ix])].reshape(1, -1)),
                np.random.choice(ite_pool_preds[int(query_idx[ix])], size=1),
                sample_weight = np.array(pool_type_s[int(query_idx[ix])]).ravel())
        out = np.zeros(dataset["X_pool"].shape[0])
        out[query_idx] = 1
        return out

# %% ../01_models.ipynb 6
class RandomAssignmentFunction(BaseAssignmentFunction):
    def __init__(self, base_selection = 0, p = .5):
        super().__init__(base_selection)
        self.p = p

    def select_treatment(self, model, dataset, query_idx):
        return np.random.binomial(1, self.p, (query_idx.shape[0],))

class UncertaintyAssignmentFunction(BaseAssignmentFunction):
    def select_treatment(self, model, dataset, query_idx):
        preds = model.predict(X = dataset["X_pool"][query_idx, :], return_per_cf = True)
        try:
            var1 = np.var(preds["pred1"], axis=1)
            var0 = np.var(preds["pred0"], axis=1)
            ex = var1 / (var1 + var0)
            out = np.random.binomial(1,ex)
        except:
            raise ValueError("Can't access uncertainty per counterfactual")
        return out

class MajorityAssignmentFunction(BaseAssignmentFunction):
    def select_treatment(self, model, dataset, query_idx):
        if sum(dataset["t_training"]) >= dataset["t_training"].shape[0]/2:
            out = np.zeros((query_idx.shape[0],))
        else:
            out = np.ones((query_idx.shape[0],))
        return out

class PMajorityAssignmentFunction(BaseAssignmentFunction):
    def __init__(self, base_selection = 0, p = .5):
        super().__init__(base_selection)
        self.p = p

    def select_treatment(self, model, dataset, query_idx):
        if sum(dataset["t_training"]) >= dataset["t_training"].shape[0]/2:
            out = np.random.binomial(query_idx.shape[0], p = 1-self.p)
        else:
            out = np.random.binomial(query_idx.shape[0], p = self.p)
        return out



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"FitTask": "00_base.ipynb",
         "BaseITEEstimator": "00_base.ipynb",
         "BaseActiveLearner": "00_base.ipynb",
         "BaseAcquisitionFunction": "00_base.ipynb",
         "BaseAssignmentFunction": "00_base.ipynb",
         "BaseStoppingRule": "00_base.ipynb",
         "BaseDataGenerator": "00_base.ipynb",
         "RandomAcquisitionFunction": "01_models.ipynb",
         "UncertaintyAcquisitionFunction": "01_models.ipynb",
         "TypeSAcquistionFunction": "01_models.ipynb",
         "EMCMAcquisitionFunction": "01_models.ipynb",
         "RandomAssignmentFunction": "01_models.ipynb",
         "UncertaintyAssignmentFunction": "01_models.ipynb",
         "MajorityAssignmentFunction": "01_models.ipynb",
         "PMajorityAssignmentFunction": "01_models.ipynb",
         "CausalForestEstimator": "02_estimators.ipynb",
         "OPENBTITEEstimator": "02_estimators.ipynb",
         "GPEstimator": "02_estimators.ipynb",
         "get_ihdp_dict": "99_misc.ipynb"}

modules = ["base.py",
           "models.py",
           "estimators.py",
           "helper.py"]

doc_url = "https://puhazoli.github.io/asbe/"

git_url = "https://github.com/puhazoli/asbe/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 01_core.ipynb (unless otherwise specified).

__all__ = ['random_batch_sampling', 'uncertainty_batch_sampling', 'type_s_batch_sampling',
           'expected_model_change_maximization', 'ASLearner', 'estimator_type', 'ITEEstimator', 'variance_based_assf']

# Cell
import numpy as np
from .base import *
from modAL.models.base import BaseLearner
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.preprocessing import StandardScaler
from sklearn.pipeline import make_pipeline
from typing import Union, Optional, Callable
from copy import deepcopy
from pylift.eval import UpliftEval

# Cell
def random_batch_sampling(classifier, X_pool, n2, **kwargs):
    "Randomly sample a batch from a pool of unlabaled samples"
    n_samples = X_pool.shape[0]
    query_idx = np.random.choice(range(n_samples), size=n2,replace=False)
    return X_pool[query_idx], query_idx

def uncertainty_batch_sampling(classifier, X_pool, n2, **kwargs):
    "Select the top $n_2$ most uncertain units"
    ite_preds, y1_preds, y_preds = classifier.predict(X_pool, **kwargs)
    # Calculate variance based on predicted
    if y1_preds.shape[0] <= 1 or \
    len(y1_preds.shape) <= 1:
            raise Exception("Not possible to calculate uncertainty when dimensions <=1 ")
    ite_vars = np.var(classifier.estimator.y1_preds - classifier.estimator.y0_preds, axis=1)
    query_idx = np.argsort(ite_vars)[-n2:][::-1]

    return X_pool[query_idx], query_idx

def type_s_batch_sampling(classifier, X_pool, n2, **kwargs):
    "Select highest type-s"
    ite_preds, y1_preds, y_preds = classifier.predict(X_pool, **kwargs)
    prob_s = np.sum(ite_preds > 0, axis=1)/ite_preds.shape[1]
    prob_s_sel = np.where(prob_s > 0.5, 1-prob_s, prob_s) + .0001
    query_idx = np.argsort(prob_s_sel)[-n2:][::-1]

    return X_pool[query_idx], query_idx


def expected_model_change_maximization(classifier, X_pool, n2, **kwargs):
    """
    Implementation of EMCM for ITE - using a surrogate SGD model.
    """
    # Get mean of the trained prediction
    ite_train_preds, y1_train_preds, y0_train_preds = \
        classifier.predict(classifier.X_training, **kwargs)
    if ite_train_preds.shape[1] < 1:
        raise ValueError("The treatment effect does not have uncertainty around it - \
                         consider using a different estimator")
    # Get mean of predicted ITE
    ite_pool_preds, y1_pool_preds, y0_pool_preds = \
        classifier.predict(X_pool, **kwargs)
    # Then scale the data so sgd works the best
    sc = StandardScaler()
    X_scaled = sc.fit_transform(classifier.X_training)
    # Fit approx model
    # calc type-s error
    train_type_s_prob_1 = np.sum(ite_train_preds > 0, axis=1)/ite_train_preds.shape[1]
    train_type_s = np.where(train_type_s_prob_1 > 0.5, 1-train_type_s_prob_1, train_type_s_prob_1) + .0001
    pool_type_s_prob_1 = np.sum(ite_pool_preds > 0, axis=1)/ite_pool_preds.shape[1]
    pool_type_s = np.where(pool_type_s_prob_1 > 0.5, 1-pool_type_s_prob_1, pool_type_s_prob_1) + .0001
    classifier.approx_model.fit(
        X = X_scaled,
        y = np.mean(ite_train_preds, axis=1),
        sample_weight = 5*train_type_s)
    # Using list as it is faster than appending to np array
    query_idx = []
    # Using a loop for the combinatorial opt. part
    for ix in range(n2):
        if n2 > (X_pool.shape[0]):
            raise IndexError("Too many samples are queried from the pool ($n_2 > ||X_pool||$)")
        # Select randomly from X_pool
        prob_sampling = np.ones((X_pool.shape[0]))/(X_pool.shape[0]-len(query_idx))
        # Set the probability of already selected samples to 0
        if ix > 0:
            prob_sampling[query_idx] = 0
        # B = 100 by default, can be modified by kwargs
        considered_ixes = np.random.choice(X_pool.shape[0],
                                         size = kwargs["B"] if "B" in kwargs else 100,
                                         replace=False,
                                         p=prob_sampling)
        # Calculate the grads for all the
        grads = np.array([])
        for considered_ix in considered_ixes:
            new_X = sc.transform(X_pool[considered_ix].reshape(1, -1))
            app_predicted_ite = classifier.approx_model.predict(new_X)
            # bootstrapping accroding to eq. 11 of Cai and Zhang
            true_ite = np.random.choice(ite_pool_preds[considered_ix],
                                        size=kwargs["K"] if "K" in kwargs else 5)
            grad = np.sum(np.abs(np.kron((true_ite - app_predicted_ite),new_X)))
            grads = np.append(grads, grad)
        if np.max(grads) < kwargs["threshold"] if "threshold" in kwargs else 0:
            break
        classifier.model_change = np.append(classifier.model_change,np.max(grads))
        query_idx.append(int(considered_ixes[np.argmax(grads)]))
        classifier.approx_model.partial_fit(
            sc.transform(X_pool[int(query_idx[ix])].reshape(1, -1)),
            np.random.choice(ite_pool_preds[int(query_idx[ix])], size=1),
            sample_weight = np.array(pool_type_s[int(query_idx[ix])]).ravel())

    return X_pool[query_idx], query_idx

# Cell
estimator_type = ClassifierMixin
class ASLearner(BaseLearner):
    """A(ctively)S(topping)Learner class for automatic stopping in batch-mode AL"""
    def __init__(self,
                 estimator: estimator_type=None,
                 query_strategy=None,
                 assignment_fc=None,
                 X_training: np.ndarray = None,
                 t_training: np.ndarray = None,
                 y_training: np.ndarray = None,
                 X_pool: np.ndarray = None,
                 X_test: np.ndarray = None,
                 approx_model: RegressorMixin = None
                ) -> None:
        self.estimator = estimator
        self.query_strategy = query_strategy
        self.assignment_fc = assignment_fc
        self.X_training = X_training
        self.y_training = y_training
        self.t_training = t_training
        self.X_pool     = X_pool
        self.X_test     = X_test
        self.approx_model = approx_model
        self.model_change = np.array([])

    def _add_queried_data_class(self, X, t, y):
        self.X_training = np.vstack((self.X_training, X))
        self.t_training = np.concatenate((self.t_training, t))
        self.y_training = np.concatenate((self.y_training, y))

    def _update_estimator_values(self):
        self.estimator.__dict__.update(X_training = self.X_training,
                               y_training  =        self.y_training,
                               t_training  =        self.t_training,
                               X_test      =        self.X_test)

    def teach(self, X_new, t_new, y_new, **kwargs):
        """Teaching new instances to the estimator selected bu the query_strategy

        If no `assignment_fc` is added, all selected samples are used
        If assignment function is added, only those instances are used, where
        $\hat{T} = T$
        """
        if self.assignment_fc is not None:
            X_new, t_new, y_new = self.assignment_fc(
                self, X_new, t_new,
                y_new, simulated=kwargs["simulated"] if "simulated" in kwargs else False)
        else:
            try:
                y_new = np.take_along_axis(y_new, t_new[:, None], axis=1)
            except:
                pass
        self._add_queried_data_class(X_new, t_new.ravel(), y_new.ravel())
        self.fit()

    def fit(self):
        self._update_estimator_values()
        self.estimator.fit()

    def predict(self, X=None, **kwargs):
        """Method for predicting treatment effects within Active Learning

        Default is to predict on the unlabeled pool"""
        if X is None:
            raise Exception("You need to supply an unlabeled pool of instances (with shape (-1,{}))".format(self.X_training.shape[1]))
        self.preds = self.estimator.predict(X, **kwargs)
        return self.preds

    def score(self, preds=None, y_true=None, t_true=None, metric = "Qini"):
        """
        Scoring the predictions - either ITE or observed outcomes are needed.

        If observed outcomes are provided, the accompanying treatments are also needed.
        """
        if metric not in ["Qini", "PEHE", "Cgains"]:
            raise ValueError(f"Please use a valid error (PEHE, Qini, Cgains), {metric} is not valid")
        if metric == "Qini":
            upev = UpliftEval(t_true, y_true, self.preds[0] if preds is None else preds)
            self.scores = upev
            vscore = self.scores.q1_aqini
        if metric == "PEHE":
            vscore = np.sqrt(np.mean(np.square(preds - y_true)))
        if metric == "Cgains":
            upev = UpliftEval(t_true, y_true, self.preds[0] if preds is None else preds)
            self.scores = upev
            vscore = self.scores.cgains
        return vscore

# Cell
class ITEEstimator(BaseEstimator):
    """ Class for building a naive estimator for ITE estimation
    """
    def __init__(self,
                 model: estimator_type = None,
                 two_model: bool = False,
                 ps: Callable = None,
                 **kwargs
                ) -> None:
        self.model = model
        self.two_model = two_model
        self.ps_model = ps

    def _fit_ps_model(self):
        if self.ps_model is not None:
            self.ps_model.fit(self.X_training, self.t_training)

    def fit(self,X_training: np.ndarray = None,
                 t_training: np.ndarray = None,
                 y_training: np.ndarray = None,
                 X_test: np.ndarray = None,
                 ps_scores: np.ndarray = None):

        if X_training is not None:
            self.X_training = X_training
            self.y_training = y_training
            self.t_training = t_training
            self.X_test = X_test
        self.N_training = self.X_training.shape[0]
        try:
            self._fit_ps_model()
            ps_scores = self.ps_model.predict_proba(self.X_training)
        except:
            ps_scores = None
            # if "N_training" not in self.__dict__:
        #     self.N_training = self.X_training.shape[0]
        if ps_scores is not None:
            X_to_train_on = np.hstack((self.X_training, ps_scores[:,1].reshape((-1, 1))))
        else:
            X_to_train_on = self.X_training
        if self.two_model:
            if hasattr(self, "m1") is False:
                self.m1 = deepcopy(self.model)
            control_ix = np.where(self.t_training == 0)[0]
            self.model.fit(X_to_train_on[control_ix,:],
                           self.y_training[control_ix])
            self.m1.fit(X_to_train_on[-control_ix,:],
                        self.y_training[-control_ix])
        else:
            self.model.fit(np.hstack((X_to_train_on,
                                      self.t_training.reshape((self.N_training, -1)))),
                           self.y_training)

    def _predict_without_proba(self, model, X, **kwargs):
        return model.predict(X,
            return_mean = kwargs["return_mean"] if "return_mean" in kwargs else True)

    def _fix_dim_pred(self, preds):
        pred_length = preds.shape[0]
        if preds.shape[1] == 1:
            if np.all(preds == 0):
                preds = np.hstack((preds, np.ones(pred_length).reshape((-1,1))))
            elif np.all(preds == 1):
                preds = np.hstack((preds, np.zeros(pred_length).reshape((-1,1))))
            preds = preds[:, ]
        return preds

    def predict(self, X=None, **kwargs):
        if X is None:
            X = self.X_test
        if self.ps_model is not None and self.ps_model.coef_ is not None:
            pred_ps_scores = self.ps_model.predict_proba(X)[:, 1]
            X = np.hstack((X, pred_ps_scores.reshape(-1, 1)))
        N_test = X.shape[0]
        try:
            if self.two_model:
                self.y1_preds = self.m1.predict_proba(X)
                self.y0_preds = self.model.predict_proba(X)
            else:
                self.y1_preds = self.model.predict_proba(
                                    np.hstack((X,
                                    np.ones(N_test).reshape(-1,1))))
                self.y0_preds = self.model.predict_proba(
                    np.hstack((X,
                               np.zeros(N_test).reshape(-1,1))))
            self.y1_preds = self._fix_dim_pred( self.y1_preds)
            self.y0_preds = self._fix_dim_pred( self.y0_preds)
        except AttributeError:
            try:
                if self.two_model:
                    self.y1_preds = self._predict_without_proba(self.m1, X, **kwargs)
                    self.y0_preds = self._predict_without_proba(self.model, X, **kwargs)
                else:
                    self.y1_preds = self._predict_without_proba(self.model,
                             np.hstack((X,
                             np.ones(N_test).reshape(-1,1))), **kwargs)
                    self.y0_preds = self._predict_without_proba(self.model,
                        np.hstack((X,
                                   np.zeros(N_test).reshape(-1,1))), **kwargs)
            except:
                raise AttributeError("No method found for predicting with the supplied class")
        return self.y1_preds - self.y0_preds, self.y1_preds, self.y0_preds

# Cell
def variance_based_assf(classifier, X, t, y, simulated=False):
    ite_preds, y1_preds, y0_preds = classifier.predict(X, return_mean=False)
    if len(y1_preds.shape) <= 1:
            raise ValueError("Not possible to calculate variance with dim {}".format(y1_preds.shape))
    prop_score = np.var(y1_preds,axis=1)/(
        np.var(y1_preds, axis=1)+np.var(y0_preds,axis=1))
    t_assigned = np.random.binomial(1, prop_score)
    if simulated:
        try:
            y = np.take_along_axis(y, t_assigned[:, None], axis=1)
            t = t_assigned
            usable_units = np.repeat(True, repeats=X.shape[0])
        except:
            raise ValueError("Potential outcomes are needed in a matrix with shape (n,2)")
    else:
        usable_units = np.where(t_assigned == t)
    return X[usable_units], t[usable_units], y[usable_units]


# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/asbe/',
                'doc_host': 'https://puhazoli.github.io',
                'git_url': 'https://github.com/puhazoli/asbe/',
                'lib_path': 'asbe'},
  'syms': { 'asbe.base': { 'asbe.base.BaseAcquisitionFunction': ('base.html#baseacquisitionfunction', 'asbe/base.py'),
                           'asbe.base.BaseAcquisitionFunction.__init__': ('base.html#baseacquisitionfunction.__init__', 'asbe/base.py'),
                           'asbe.base.BaseAcquisitionFunction.calculate_metrics': ( 'base.html#baseacquisitionfunction.calculate_metrics',
                                                                                    'asbe/base.py'),
                           'asbe.base.BaseAcquisitionFunction.select_data': ( 'base.html#baseacquisitionfunction.select_data',
                                                                              'asbe/base.py'),
                           'asbe.base.BaseActiveLearner': ('base.html#baseactivelearner', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.__init__': ('base.html#baseactivelearner.__init__', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner._select_counterfactuals': ( 'base.html#baseactivelearner._select_counterfactuals',
                                                                                    'asbe/base.py'),
                           'asbe.base.BaseActiveLearner._update_dataset': ('base.html#baseactivelearner._update_dataset', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.fit': ('base.html#baseactivelearner.fit', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.plot': ('base.html#baseactivelearner.plot', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.predict': ('base.html#baseactivelearner.predict', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.query': ('base.html#baseactivelearner.query', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.score': ('base.html#baseactivelearner.score', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.simulate': ('base.html#baseactivelearner.simulate', 'asbe/base.py'),
                           'asbe.base.BaseActiveLearner.teach': ('base.html#baseactivelearner.teach', 'asbe/base.py'),
                           'asbe.base.BaseAssignmentFunction': ('base.html#baseassignmentfunction', 'asbe/base.py'),
                           'asbe.base.BaseAssignmentFunction.__init__': ('base.html#baseassignmentfunction.__init__', 'asbe/base.py'),
                           'asbe.base.BaseAssignmentFunction.select_treatment': ( 'base.html#baseassignmentfunction.select_treatment',
                                                                                  'asbe/base.py'),
                           'asbe.base.BaseDataGenerator': ('base.html#basedatagenerator', 'asbe/base.py'),
                           'asbe.base.BaseDataGenerator.__getitem__': ('base.html#basedatagenerator.__getitem__', 'asbe/base.py'),
                           'asbe.base.BaseDataGenerator.get_X': ('base.html#basedatagenerator.get_x', 'asbe/base.py'),
                           'asbe.base.BaseDataGenerator.get_data': ('base.html#basedatagenerator.get_data', 'asbe/base.py'),
                           'asbe.base.BaseDataGenerator.get_t': ('base.html#basedatagenerator.get_t', 'asbe/base.py'),
                           'asbe.base.BaseDataGenerator.get_y': ('base.html#basedatagenerator.get_y', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator': ('base.html#baseiteestimator', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator.__init__': ('base.html#baseiteestimator.__init__', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator._fit_ps_model': ('base.html#baseiteestimator._fit_ps_model', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator._predict_bin_or_con': ( 'base.html#baseiteestimator._predict_bin_or_con',
                                                                               'asbe/base.py'),
                           'asbe.base.BaseITEEstimator.fit': ('base.html#baseiteestimator.fit', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator.predict': ('base.html#baseiteestimator.predict', 'asbe/base.py'),
                           'asbe.base.BaseITEEstimator.prepare_data': ('base.html#baseiteestimator.prepare_data', 'asbe/base.py'),
                           'asbe.base.BaseStoppingRule': ('base.html#basestoppingrule', 'asbe/base.py'),
                           'asbe.base.BaseStoppingRule.__init__': ('base.html#basestoppingrule.__init__', 'asbe/base.py'),
                           'asbe.base.BaseStoppingRule.check_rule': ('base.html#basestoppingrule.check_rule', 'asbe/base.py'),
                           'asbe.base.FitTask': ('base.html#fittask', 'asbe/base.py'),
                           'asbe.base.FitTask.__init__': ('base.html#fittask.__init__', 'asbe/base.py')},
            'asbe.core': {},
            'asbe.estimators': { 'asbe.estimators.CausalForestEstimator': ('estimators.html#causalforestestimator', 'asbe/estimators.py'),
                                 'asbe.estimators.CausalForestEstimator.fit': ( 'estimators.html#causalforestestimator.fit',
                                                                                'asbe/estimators.py'),
                                 'asbe.estimators.CausalForestEstimator.predict': ( 'estimators.html#causalforestestimator.predict',
                                                                                    'asbe/estimators.py'),
                                 'asbe.estimators.GPEstimator': ('estimators.html#gpestimator', 'asbe/estimators.py'),
                                 'asbe.estimators.GPEstimator.predict': ('estimators.html#gpestimator.predict', 'asbe/estimators.py'),
                                 'asbe.estimators.OPENBTITEEstimator': ('estimators.html#openbtiteestimator', 'asbe/estimators.py'),
                                 'asbe.estimators.OPENBTITEEstimator.predict': ( 'estimators.html#openbtiteestimator.predict',
                                                                                 'asbe/estimators.py'),
                                 'asbe.estimators.SKLiftEstimator': ('estimators.html#skliftestimator', 'asbe/estimators.py'),
                                 'asbe.estimators.SKLiftEstimator.__init__': ( 'estimators.html#skliftestimator.__init__',
                                                                               'asbe/estimators.py'),
                                 'asbe.estimators.SKLiftEstimator.fit': ('estimators.html#skliftestimator.fit', 'asbe/estimators.py'),
                                 'asbe.estimators.SKLiftEstimator.predict': ( 'estimators.html#skliftestimator.predict',
                                                                              'asbe/estimators.py')},
            'asbe.helper': {'asbe.helper.get_ihdp_dict': ('misc.html#get_ihdp_dict', 'asbe/helper.py')},
            'asbe.models': { 'asbe.models.EMCMAcquisitionFunction': ('models.html#emcmacquisitionfunction', 'asbe/models.py'),
                             'asbe.models.EMCMAcquisitionFunction.__init__': ( 'models.html#emcmacquisitionfunction.__init__',
                                                                               'asbe/models.py'),
                             'asbe.models.EMCMAcquisitionFunction.calculate_metrics': ( 'models.html#emcmacquisitionfunction.calculate_metrics',
                                                                                        'asbe/models.py'),
                             'asbe.models.MajorityAssignmentFunction': ('models.html#majorityassignmentfunction', 'asbe/models.py'),
                             'asbe.models.MajorityAssignmentFunction.select_treatment': ( 'models.html#majorityassignmentfunction.select_treatment',
                                                                                          'asbe/models.py'),
                             'asbe.models.PMajorityAssignmentFunction': ('models.html#pmajorityassignmentfunction', 'asbe/models.py'),
                             'asbe.models.PMajorityAssignmentFunction.__init__': ( 'models.html#pmajorityassignmentfunction.__init__',
                                                                                   'asbe/models.py'),
                             'asbe.models.PMajorityAssignmentFunction.select_treatment': ( 'models.html#pmajorityassignmentfunction.select_treatment',
                                                                                           'asbe/models.py'),
                             'asbe.models.RandomAcquisitionFunction': ('models.html#randomacquisitionfunction', 'asbe/models.py'),
                             'asbe.models.RandomAcquisitionFunction.calculate_metrics': ( 'models.html#randomacquisitionfunction.calculate_metrics',
                                                                                          'asbe/models.py'),
                             'asbe.models.RandomAssignmentFunction': ('models.html#randomassignmentfunction', 'asbe/models.py'),
                             'asbe.models.RandomAssignmentFunction.__init__': ( 'models.html#randomassignmentfunction.__init__',
                                                                                'asbe/models.py'),
                             'asbe.models.RandomAssignmentFunction.select_treatment': ( 'models.html#randomassignmentfunction.select_treatment',
                                                                                        'asbe/models.py'),
                             'asbe.models.TypeSAcquistionFunction': ('models.html#typesacquistionfunction', 'asbe/models.py'),
                             'asbe.models.TypeSAcquistionFunction.calculate_metrics': ( 'models.html#typesacquistionfunction.calculate_metrics',
                                                                                        'asbe/models.py'),
                             'asbe.models.UncertaintyAcquisitionFunction': ('models.html#uncertaintyacquisitionfunction', 'asbe/models.py'),
                             'asbe.models.UncertaintyAcquisitionFunction.calculate_metrics': ( 'models.html#uncertaintyacquisitionfunction.calculate_metrics',
                                                                                               'asbe/models.py'),
                             'asbe.models.UncertaintyAssignmentFunction': ('models.html#uncertaintyassignmentfunction', 'asbe/models.py'),
                             'asbe.models.UncertaintyAssignmentFunction.select_treatment': ( 'models.html#uncertaintyassignmentfunction.select_treatment',
                                                                                             'asbe/models.py')}}}
